/*
 *  Copyright (C) 2002-2006  The DOSBox Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* $Id: setup.h,v 1.22 2006/02/09 11:47:48 qbix79 Exp $ */

#ifndef DOSBOX_SETUP_H
#define DOSBOX_SETUP_H

#ifdef _MSC_VER
#pragma warning ( disable : 4786 )
#endif

#ifndef DOSBOX_CROSS_H
#include "cross.h"
#endif
#define MAX_CMDS 64
class CommandLine {
public:
	CommandLine(int argc,char * argv[]);
	CommandLine(char * name,char * cmdline);
	const char * GetFileName(){ return file_name;}	

	bool FindExist(char * name,bool remove=false);
	bool FindInt(char * name,int & value,bool remove=false);
	bool FindString(char * name,char* & value,bool remove=false);
	bool FindCommand(unsigned int which,char* & value);
	bool FindStringRemain(char * name,char* & value);
	bool GetStringRemain(char* & value);
	unsigned int GetCount(void);
//private:
	typedef char** cmd_it;
	class commands{
	public:
		commands(){count=0;}
		int count;
		char* komande[MAX_CMDS];
		void push_back(char *stri){
			if(count>=MAX_CMDS) return;
			komande[count]=stri;
//			printf("begin:%s:%s:%d\n", komande[0],stri,count);
			count++;
		}
		void erase(char **stri){
			cmd_it cnt;
			if ((stri>=end())||(stri<begin())) return;
			for(cnt=stri;cnt<end();cnt++) *cnt=*(cnt+1);
			count--;
		}
		void erase(char **first,char **last){
			cmd_it cnt;
			int howmuch=(last-first)+1;
			if(howmuch<1) return;
			for(cnt=first;cnt<=last;cnt++){
				*cnt=*(cnt+howmuch);
			}
			count-=howmuch;
		}
		int size(){
			return count;
		}
		char **begin(){
			return komande;
		}
		char **end(){
			return (komande+count);
		}
	}cmds;
	char* file_name;
	bool FindEntry(char * name,cmd_it &it,bool neednext=false);
};

class std_string{
public:
	char *str;
	int length;
	std_string() {
		length=18;
		str=(char*)malloc(20);
		str[0]='\0';
	}
	std_string(const char *stri) {
		length=strlen(stri);
		str=(char*)malloc(length+1);
		strcpy(str,stri);
		str[length]='\0';
	}
	~std_string() {if(str) free(str);}
	int size (){
		return strlen(str);
	}
	const char* c_str (){
		return str;
	}
	void assign(const char* strnew){
		if(strlen(strnew)>length){
			free(str);
			length=strlen(strnew);
			str=(char*)malloc(length+1);
			strcpy(str,strnew);
		}
		else strcpy(str,strnew);
	}
	void Strcat(const char* strc){
		int newlen=strlen(str)+strlen(strc);
		if(newlen>=length){
			char *tmpstr;
			tmpstr=(char*)malloc(newlen+2);
			length=newlen;
			strcpy(tmpstr,str);
			strcat(tmpstr,strc);
			free(str);
			str=tmpstr;
		}
		else strcat(str,strc);
	}
	void Strcat(const char strc){
		int newlen=strlen(str)+1;
		if(newlen>=length){
			char *tmpstr;
			tmpstr=(char*)malloc(newlen+2);
			length=newlen;
			strcpy(tmpstr,str);
			tmpstr[strlen(str)+1]='\0';
			tmpstr[strlen(str)]=strc;
			free(str);
			str=tmpstr;
		}
		else {
			int i=strlen(str);
			str[i+1]='\0';
			str[i]=strc;
		}
	}
	std_string* erase(){
		if(!str) return NULL;
		str[0]='\0';
		return this;
	}
	std_string* erase(int start,int siz){
		if(!str) return NULL;
		if(start>=size()) return this;
		*(str+start)='\0';
		return this;
	}
	char* begin(){
		return str;
	}
	char* end(){
		return str+strlen(str);
	}
};

union Value{
	int _hex;
	bool _bool;
	int _int;
	std_string* _string;
	float _float;   
};

class Property {
public:
	Property(const char* _propname):propname(_propname) { }
	virtual void SetValue(char* input)=0;
	virtual void GetValuestring(char* str)=0;
	Value GetValue() { return value;}
	virtual ~Property(){ }
	const char * propname;
	Value value;
};

class Prop_int:public Property {
public:
	Prop_int(const char* _propname, int _value):Property(_propname) { 
		value._int=_value;
	}
	void SetValue(char* input);
        void GetValuestring(char* str);
	~Prop_int(){ }
};
class Prop_float:public Property {
public:
	Prop_float(const char* _propname, float _value):Property(_propname){
		value._float=_value;
	}
	void SetValue(char* input);
	void GetValuestring(char* str);
	~Prop_float(){ }
};

class Prop_bool:public Property {
public:
	Prop_bool(const char* _propname, bool _value):Property(_propname) { 
		value._bool=_value;
	}
	void SetValue(char* input);
        void GetValuestring(char* str);
	~Prop_bool(){ }
};

class Prop_string:public Property{
public:
	Prop_string(const char* _propname, char* _value):Property(_propname) { 
		value._string=new std_string(_value);
	}
	~Prop_string(){
		delete value._string;
	}
	void SetValue(char* input);
        void GetValuestring(char* str);
};
class Prop_hex:public Property {
public:
	Prop_hex(const char* _propname, int _value):Property(_propname) { 
		value._hex=_value;
	}
	void SetValue(char* input);
	~Prop_hex(){ }
	void GetValuestring(char* str);
};

typedef void (*SectionFunction)(Section*);
class Function_wrapper {
public:
	Function_wrapper(SectionFunction _fun,bool _ch){
		function=_fun;
		canchange=_ch;
	}
	SectionFunction function;
	bool canchange;
};

class Section {
private:
	/* Wrapper class around startup and shutdown functions. the variable
	 * canchange indicates it can be called on configuration changes */
	struct initf{
		Function_wrapper* funcs[50];
		int count;
		void push_back(Function_wrapper *fw){
			funcs[count]=fw;
			count++;
		}
		Function_wrapper** begin(){
			return funcs;
		}
		Function_wrapper** end(){
			return funcs+count;
		}
	}initfunctions;
	struct destf{
		Function_wrapper* funcs[50];
		int count;
		void push_front(Function_wrapper *fw){
			int i;
			for(i=count;i>0;i--) funcs[i]=funcs[i-1];
			funcs[0]=fw;
			count++;
		}
		Function_wrapper** erase(Function_wrapper** fw){
			int i,j;
			if(!count) return funcs;
			if(count==1) {
				count--;
				free(funcs[0]);
				return funcs;
			}
			bool found=false;
			int which=0;
			for(i=0;i<count;i++){
				if((!found)&&((funcs+i)==fw)) {
					found=true;
					which=i;
					free(funcs[i]);
				}
				if(found){
					funcs[i]=funcs[i+1];
				}
			}
			if(!found) return (funcs+count);
			count--;
			return (funcs+which);
		}
		Function_wrapper** begin(){
			return funcs;
		}
		Function_wrapper** end(){
			return funcs+count;
		}
	}destroyfunctions;
	char sectionname[20];
public:
	Section(const char* _sectionname) { initfunctions.count=0; destroyfunctions.count=0; strcpy(sectionname,_sectionname);}

	void AddInitFunction(SectionFunction func,bool canchange=false) {initfunctions.push_back(new Function_wrapper(func,canchange));}
	void AddDestroyFunction(SectionFunction func,bool canchange=false) {destroyfunctions.push_front(new Function_wrapper(func,canchange));}
	void ExecuteInit(bool initall=true);
	void ExecuteDestroy(bool destroyall=true);
	const char* GetName() {return sectionname;}

	virtual char* GetPropValue(const char* _property)=0;
	virtual void HandleInputline(char * _line)=0;
//	virtual void PrintData(FILE* outfile)=0;
	virtual ~Section() {
		for(int i=0;i<initfunctions.count;i++) free(initfunctions.funcs[i]);
		for(int i=0;i<destroyfunctions.count;i++) free(destroyfunctions.funcs[i]);
	}
};


class Section_prop:public Section {
private:
//	std::list<Property*> properties;
	typedef Property** it;
public:
	struct sekcija{
		int count;
		Property* sects[50];
		Property **push_back(Property* sec){
			sects[count]=sec;
			count++;
		}

		Property **begin() {return sects;}
		Property **end() {return sects+count;}
	}properties;
	Section_prop(const char* _sectionname):Section(_sectionname){properties.count=0;}
	void Add_int(const char* _propname, int _value=0);
	void Add_string(const char* _propname, char* _value=NULL);
	void Add_bool(const char* _propname, bool _value=false);
	void Add_hex(const char* _propname, int _value=0);
	void Add_float(const char* _propname, float _value=0.0);   

	int Get_int(const char* _propname);
	const char* Get_string(const char* _propname);
	bool Get_bool(const char* _propname);
	int Get_hex(const char* _propname);
	float Get_float(const char* _propname);
	void HandleInputline(char *gegevens);
	void PrintData(FILE* outfile);
	virtual char* GetPropValue(const char* _property);
	//ExecuteDestroy should be here else the destroy functions use destroyed properties
	virtual ~Section_prop();
};

class Section_line: public Section{
public:
	Section_line(const char* _sectionname):Section(_sectionname){data=new std_string("");}
	~Section_line(){delete data;ExecuteDestroy(true);}
	void HandleInputline(char* gegevens);
	void PrintData(FILE* outfile);
	virtual char* GetPropValue(const char* _property);
	std_string* data;
};

class Config{
public:
	CommandLine * cmdline;
private:
	typedef Section** it;
	void (* _start_function)(void);
public:
	Config(CommandLine * cmd){ cmdline=cmd; sectionlist.count=0;}
	~Config();
	struct sekcija{
		int count;
		Section* sects[50];
		Section **push_back(Section* sec){
			sects[count]=sec;
			count++;
		}

		Section **begin() {return sects;}
		Section **end() {return sects+count;}
	}sectionlist;
	Section_line * AddSection_line(const char * _name,void (*_initfunction)(Section*));
	Section_prop * AddSection_prop(const char * _name,void (*_initfunction)(Section*),bool canchange=false);
	
	Section* GetSection(const char* _sectionname);
	Section* GetSectionFromProperty(const char* prop);

	void SetStartUp(void (*_function)(void));
	void Init();
	void ShutDown();
	void StartUp();
	void PrintConfig(const char* configfilename);
	bool ParseConfigFile(const char* configfilename);
	void ParseEnv(char ** envp);
};

class Module_base {
	/* Base for all hardware and software "devices" */
protected:
	Section* m_configuration;
public:
	Module_base(Section* configuration){m_configuration=configuration;};
//	Module_base(Section* configuration, SaveState* state) {};
	~Module_base(){/*LOG_MSG("executed")*/;};//Destructors are required
	/* Returns true if succesful.*/
	virtual bool Change_Config(Section* newconfig) {return false;} ;
};
#endif
