#define ALLOW_ACCESS_TO_INTERNALS_OF_BITMAPS

#include <PalmTypes.h>
#include <PceNativeCall.h>
#include <CoreTraps.h>
#include <VFSMgr.h>
#include <TimeMgr.h>
#include <StringMgr.h>
#include <DateTime.h>
#include "PalmOSfunc.h"
#include <SysUtils.h>
#include <Bitmap.h>
//     #include <stdio.h>
     #include <stdarg.h>
//#include <sys/stat.h>

extern void *gEmulStateP;
extern Call68KFuncType *gCall68KFuncP;

unsigned long
Byte_Swap32( unsigned long l )
{
    unsigned long result;
    result  = ((l >> 24) & 0xFF);
    result |= ((l >> 8) & 0xFF00);
    result |= ((l << 8) & 0xFF0000);
    result |= ((l << 24) & 0xFF000000);
    return result;
}

unsigned short
Byte_Swap16( unsigned short l )
{
    unsigned short result;
    result  = ((l >> 8) & 0xFF);
    result |= ((l << 8) & 0xFF00);
    return result;
}

void
write_unaligned32( unsigned char* dest, unsigned long val )
{
    int i;
    dest += sizeof(val);
    for ( i = 0; i < sizeof(val); ++i ) {
        *--dest = val & 0x000000FF;
        val >>= 8;
    }
} /* write_unaligned32 */

void
write_unaligned16( unsigned char* dest, unsigned short val )
{
    int i;

    dest += sizeof(val);
    for ( i = 0; i < sizeof(val); ++i ) {
        *--dest = val & 0x00FF;
        val >>= 8;
    }
} /* write_unaligned16 */

#define write_unaligned8( p, v ) *(p) = v

unsigned short 
read_unaligned16( const unsigned char* src )
{
    int i;
    unsigned short val = 0;

    for ( i = 0; i < sizeof(val); ++i ) {
        val <<= 8;
        val |= *src++;
    }

    return val;
} /* read_unaligned16 */

#define read_unaligned8(cp) (*(cp))

unsigned long
read_unaligned32( const unsigned char* src )
{
    int i;
    unsigned long val = 0;

    for ( i = 0; i < sizeof(val); ++i ) {
        val <<= 8;
        val |= *src++;
    }

    return val;
} /* read_unaligned32 */

#define STACK_START(typ, var, siz ) typ var[siz]
# define FUNC_HEADER(n)
# define FUNC_TAIL(n)
#define STACK_END(s)
#define ADD_TO_STACK4(s,val,offset) \
    write_unaligned32( &s[offset], (unsigned long)val )
#define ADD_TO_STACK2(s,val,offset) \
    write_unaligned16( &s[offset], (unsigned short)val )
#define ADD_TO_STACK1(s,val,offset) \
    s[offset] = (unsigned char)val; \
    s[offset+1] = 0
#define SWAP1_NON_NULL_IN(ptr) if ( !!(ptr) ) { *(ptr) = Byte_Swap16(*(ptr)); }
#define SWAP2_NON_NULL_IN(ptr) if ( !!(ptr) ) { *(ptr) = Byte_Swap16(*(ptr)); }
#define SWAP4_NON_NULL_IN(ptr) if ( !!(ptr) ) { *(ptr) = Byte_Swap32(*(ptr)); }
#define SWAP1_NON_NULL_OUT(ptr) SWAP1_NON_NULL_IN(ptr) 
#define SWAP2_NON_NULL_OUT(ptr) SWAP2_NON_NULL_IN(ptr) 
#define SWAP4_NON_NULL_OUT(ptr) SWAP4_NON_NULL_IN(ptr) 
# define EMIT_NAME(n,b)

void callm68k(void *funcP,void *rtoP){
         gCall68KFuncP( gEmulStateP, (unsigned long)funcP, rtoP, kPceNativeWantA0 | 4 );
}

Char*
StrCopy( Char* dst, const Char* src )
{
    Char* result;
    FUNC_HEADER(StrCopy);
    /* var decls */
    /* swapIns */
    {
    STACK_START(unsigned char, stack, 8);
    /* pushes */
    ADD_TO_STACK4(stack, dst, 0);
    ADD_TO_STACK4(stack, src, 4);
    STACK_END(stack);
    result = (Char*)gCall68KFuncP( gEmulStateP, 
                               PceNativeTrapNo(sysTrapStrCopy),
                               stack, kPceNativeWantA0 | 8 );
    /* swapOuts */
    }
    FUNC_TAIL(StrCopy);
    EMIT_NAME("StrCopy","'S','t','r','C','o','p','y'");
    return result;
} /* StrCopy */

Int16
StrNCompare( const Char* s1, const Char* s2, Int32 n )
{
    Int16 result;
    FUNC_HEADER(StrNCompare);
    /* var decls */
    /* swapIns */
    {
    STACK_START(unsigned char, stack, 12);
    /* pushes */
    ADD_TO_STACK4(stack, s1, 0);
    ADD_TO_STACK4(stack, s2, 4);
    ADD_TO_STACK4(stack, n, 8);
    STACK_END(stack);
    result = (Int16)gCall68KFuncP( gEmulStateP, 
                               PceNativeTrapNo(sysTrapStrNCompare),
                               stack, 12 );
    /* swapOuts */
    }
    FUNC_TAIL(StrNCompare);
    EMIT_NAME("StrNCompare","'S','t','r','N','C','o','m','p','a','r','e'");
    return result;
} /* StrNCompare */

Int16
StrCompare( const Char* s1, const Char* s2 )
{
    Int16 result;
    FUNC_HEADER(StrCompare);
    /* var decls */
    /* swapIns */
    {
    STACK_START(unsigned char, stack, 8);
    /* pushes */
    ADD_TO_STACK4(stack, s1, 0);
    ADD_TO_STACK4(stack, s2, 4);
    STACK_END(stack);
    result = (Int16)gCall68KFuncP( gEmulStateP, 
                               PceNativeTrapNo(sysTrapStrCompare),
                               stack, 8 );
    /* swapOuts */
    }
    FUNC_TAIL(StrCompare);
    EMIT_NAME("StrCompare","'S','t','r','C','o','m','p','a','r','e'");
    return result;
} /* StrCompare */

Char*
StrCat( Char* dst, const Char* src )
{
    Char* result;
    FUNC_HEADER(StrCat);
    /* var decls */
    /* swapIns */
    {
    STACK_START(unsigned char, stack, 8);
    /* pushes */
    ADD_TO_STACK4(stack, dst, 0);
    ADD_TO_STACK4(stack, src, 4);
    STACK_END(stack);
    result = (Char*)gCall68KFuncP( gEmulStateP, 
                               PceNativeTrapNo(sysTrapStrCat),
                               stack, kPceNativeWantA0 | 8 );
    /* swapOuts */
    }
    FUNC_TAIL(StrCat);
    EMIT_NAME("StrCat","'S','t','r','C','a','t'");
    return result;
} /* StrCat */

Err
MemMove( void* dstP, const void* sP, Int32 numBytes )
{
    Err result;
    {
    STACK_START(unsigned char, stack, 12);
    ADD_TO_STACK4(stack, dstP, 0);
    ADD_TO_STACK4(stack, sP, 4);
    ADD_TO_STACK4(stack, numBytes, 8);
    STACK_END(stack);
    result = (Err)gCall68KFuncP( gEmulStateP, 
                               PceNativeTrapNo(sysTrapMemMove),
                               stack, 12 );
    }
    return result;
}

UInt16
StrLen( const Char* src )
{
    UInt16 result;
    FUNC_HEADER(StrLen);
    /* var decls */
    /* swapIns */
    {
    STACK_START(unsigned char, stack, 4);
    /* pushes */
    ADD_TO_STACK4(stack, src, 0);
    STACK_END(stack);
    result = (UInt16)gCall68KFuncP( gEmulStateP, 
                               PceNativeTrapNo(sysTrapStrLen),
                               stack, 4 );
    /* swapOuts */
    }
    FUNC_TAIL(StrLen);
    EMIT_NAME("StrLen","'S','t','r','L','e','n'");
    return result;
} /* StrLen */

Err
MemSet( void* dstP, Int32 numBytes, UInt8 value )
{
    Err result;
    FUNC_HEADER(MemSet);
    /* var decls */
    /* swapIns */
    {
    STACK_START(unsigned char, stack, 10);
    /* pushes */
    ADD_TO_STACK4(stack, dstP, 0);
    ADD_TO_STACK4(stack, numBytes, 4);
    ADD_TO_STACK1(stack, value, 8);
    STACK_END(stack);
    result = (Err)gCall68KFuncP( gEmulStateP, 
                               PceNativeTrapNo(sysTrapMemSet),
                               stack, 10 );
    /* swapOuts */
    }
    FUNC_TAIL(MemSet);
    EMIT_NAME("MemSet","'M','e','m','S','e','t'");
    return result;
} /* MemSet */

Err
VFSFileRead( FileRef fileRef, UInt32 numBytes, void* bufP, UInt32* numBytesReadP )
{
    Err result;
    FUNC_HEADER(VFSFileRead);
    SWAP4_NON_NULL_IN(numBytesReadP);
    {
    STACK_START(unsigned char, stack, 16);
    ADD_TO_STACK4(stack, fileRef, 0);
    ADD_TO_STACK4(stack, numBytes, 4);
    ADD_TO_STACK4(stack, bufP, 8);
    ADD_TO_STACK4(stack, numBytesReadP, 12);
    STACK_END(stack);
    SET_SEL_REG(vfsTrapFileRead);
    result = (Err)gCall68KFuncP( gEmulStateP, 
                               PceNativeTrapNo(sysTrapFileSystemDispatch),
                               stack, 16 );
    SWAP4_NON_NULL_OUT(numBytesReadP);
    }
    FUNC_TAIL(VFSFileRead);
    EMIT_NAME("VFSFileRead","'V','F','S','F','i','l','e','R','e','a','d'");
    return result;
}

Err
VFSFileClose( FileRef fileRef )
{
    Err result;
    FUNC_HEADER(VFSFileClose);
    {
    STACK_START(unsigned char, stack, 4);
    ADD_TO_STACK4(stack, fileRef, 0);
    STACK_END(stack);
    SET_SEL_REG(vfsTrapFileClose);
    result = (Err)gCall68KFuncP( gEmulStateP, 
                               PceNativeTrapNo(sysTrapFileSystemDispatch),
                               stack, 4 );
    }
    FUNC_TAIL(VFSFileClose);
    EMIT_NAME("VFSFileClose","'V','F','S','F','i','l','e','C','l','o','s','e'");
    return result;
}

Err
VFSFileOpen( UInt16 volRefNum, const Char* pathNameP, UInt16 openMode, FileRef* fileRefP )
{
    Err result;
    FUNC_HEADER(VFSFileOpen);
    SWAP4_NON_NULL_IN(fileRefP);
    {
    STACK_START(unsigned char, stack, 12);
    ADD_TO_STACK2(stack, volRefNum, 0);
    ADD_TO_STACK4(stack, pathNameP, 2);
    ADD_TO_STACK2(stack, openMode, 6);
    ADD_TO_STACK4(stack, fileRefP, 8);
    STACK_END(stack);
    SET_SEL_REG(vfsTrapFileOpen);
    result = (Err)gCall68KFuncP( gEmulStateP, 
                               PceNativeTrapNo(sysTrapFileSystemDispatch),
                               stack, 12 );
    SWAP4_NON_NULL_OUT(fileRefP);
    }
    FUNC_TAIL(VFSFileOpen);
    EMIT_NAME("VFSFileOpen","'V','F','S','F','i','l','e','O','p','e','n'");
    return result;
}

Err
VFSFileCreate( UInt16 volRefNum, const Char* pathNameP )
{
    Err result;
    FUNC_HEADER(VFSFileOpen);
    {
    STACK_START(unsigned char, stack, 6);
    ADD_TO_STACK2(stack, volRefNum, 0);
    ADD_TO_STACK4(stack, pathNameP, 2);
    STACK_END(stack);
    SET_SEL_REG(vfsTrapFileCreate);
    result = (Err)gCall68KFuncP( gEmulStateP, 
                               PceNativeTrapNo(sysTrapFileSystemDispatch),
                               stack, 6 );
    }
    FUNC_TAIL(VFSFileOpen);
    EMIT_NAME("VFSFileOpen","'V','F','S','F','i','l','e','O','p','e','n'");
    return result;
}

Err
VFSFileSeek( FileRef fileRef, FileOrigin origin, Int32 offset )
{
    Err result;
    FUNC_HEADER(VFSFileSeek);
    {
    STACK_START(unsigned char, stack, 10);
    ADD_TO_STACK4(stack, fileRef, 0);
    ADD_TO_STACK2(stack, origin, 4);
    ADD_TO_STACK4(stack, offset, 6);
    STACK_END(stack);
    SET_SEL_REG(vfsTrapFileSeek);
    result = (Err)gCall68KFuncP( gEmulStateP, 
                               PceNativeTrapNo(sysTrapFileSystemDispatch),
                               stack, 10 );
    }
    FUNC_TAIL(VFSFileSeek);
    EMIT_NAME("VFSFileSeek","'V','F','S','F','i','l','e','S','e','e','k'");
    return result;
}

Err
VFSFileTell( FileRef fileRef, UInt32* filePosP )
{
    Err result;
    FUNC_HEADER(VFSFileTell);
    SWAP4_NON_NULL_IN(filePosP);
    {
    STACK_START(unsigned char, stack, 8);
    ADD_TO_STACK4(stack, fileRef, 0);
    ADD_TO_STACK4(stack, filePosP, 4);
    STACK_END(stack);
    SET_SEL_REG(vfsTrapFileTell);
    result = (Err)gCall68KFuncP( gEmulStateP, 
                               PceNativeTrapNo(sysTrapFileSystemDispatch),
                               stack, 8 );
    SWAP4_NON_NULL_OUT(filePosP);
    }
    FUNC_TAIL(VFSFileTell);
    EMIT_NAME("VFSFileTell","'V','F','S','F','i','l','e','T','e','l','l'");
    return result;
}

void SWAP_FILEINFOTYPE (char *dest, char *src){
	*dest=*(src+3);
	*(dest+1)=*(src+2);
	*(dest+2)=*(src+1);
	*(dest+3)=*src;
	*(dest+4)=*(src+7);
	*(dest+5)=*(src+6);
	*(dest+6)=*(src+5);
	*(dest+7)=*(src+4);
	*(dest+8)=*(src+9);
	*(dest+9)=*(src+8);
}

Err
VFSDirEntryEnumerate( FileRef dirRef, UInt32* dirEntryIteratorP, FileInfoType* infoP )
{
    Err result;
    FUNC_HEADER(VFSDirEntryEnumerate);
    FileInfoType FileInfoType_68K2;
    SWAP4_NON_NULL_IN(dirEntryIteratorP);
    SWAP_FILEINFOTYPE( (char*)&FileInfoType_68K2, (char*)infoP );
    {
    STACK_START(unsigned char, stack, 12);
    ADD_TO_STACK4(stack, dirRef, 0);
    ADD_TO_STACK4(stack, dirEntryIteratorP, 4);
    ADD_TO_STACK4(stack, &FileInfoType_68K2, 8);
    STACK_END(stack);
    SET_SEL_REG(vfsTrapDirEntryEnumerate);
    result = (Err)gCall68KFuncP( gEmulStateP, 
                               PceNativeTrapNo(sysTrapFileSystemDispatch),
                               stack, 12 );
    SWAP4_NON_NULL_OUT(dirEntryIteratorP);
    SWAP_FILEINFOTYPE( (char*)infoP, (char*)&FileInfoType_68K2 );
    }
    FUNC_TAIL(VFSDirEntryEnumerate);
    EMIT_NAME("VFSDirEntryEnumerate","'V','F','S','D','i','r','E','n','t','r','y','E','n','u','m','e','r','a','t','e'");
    return result;
}

Int16
MemCmp( const void* s1, const void* s2, Int32 numBytes )
{
    Int16 result;
    FUNC_HEADER(MemCmp);
    {
    STACK_START(unsigned char, stack, 12);
    ADD_TO_STACK4(stack, s1, 0);
    ADD_TO_STACK4(stack, s2, 4);
    ADD_TO_STACK4(stack, numBytes, 8);
    STACK_END(stack);
    result = (Int16)gCall68KFuncP( gEmulStateP, 
                               PceNativeTrapNo(sysTrapMemCmp),
                               stack, 12 );
    }
    FUNC_TAIL(MemCmp);
    EMIT_NAME("MemCmp","'M','e','m','C','m','p'");
    return result;
}

UInt32
TimGetSeconds(  )
{
    UInt32 result;
    FUNC_HEADER(TimGetSeconds);
    {
    STACK_START(unsigned char, stack, 0);
    STACK_END(stack);
    result = (UInt32)gCall68KFuncP( gEmulStateP, 
                               PceNativeTrapNo(sysTrapTimGetSeconds),
                               stack, 0 );
    }
    FUNC_TAIL(TimGetSeconds);
    EMIT_NAME("TimGetSeconds","'T','i','m','G','e','t','S','e','c','o','n','d','s'");
    return result;
}

UInt32
TimGetTicks(  )
{
    UInt32 result;
    FUNC_HEADER(TimGetTicks);
    {
    STACK_START(unsigned char, stack, 0);
    STACK_END(stack);
    result = (UInt32)gCall68KFuncP( gEmulStateP, 
                               PceNativeTrapNo(sysTrapTimGetTicks),
                               stack, 0 );
    }
    FUNC_TAIL(TimGetTicks);
    EMIT_NAME("TimGetTicks","'T','i','m','G','e','t','T','i','c','k','s'");
    return result;
}

void SWAP_DATETIMETYPE_68K_TO_ARM (char *dest, char *src){
	*dest=*(src+1);
	*(dest+1)=*src;
	*(dest+2)=*(src+3);
	*(dest+3)=*(src+2);
	*(dest+4)=*(src+5);
	*(dest+5)=*(src+4);
	*(dest+6)=*(src+7);
	*(dest+7)=*(src+6);
	*(dest+8)=*(src+9);
	*(dest+9)=*(src+8);
	*(dest+10)=*(src+11);
	*(dest+11)=*(src+10);
	*(dest+12)=*(src+13);
	*(dest+13)=*(src+12);
}

void
TimSecondsToDateTime( UInt32 seconds, DateTimeType* dateTimeP )
{
    FUNC_HEADER(TimSecondsToDateTime);
    DateTimeType DateTimeType_68K1;
//    SWAP_DATETIMETYPE_ARM_TO_68K( (char*)&DateTimeType_68K1, (char*)dateTimeP );
    {
    STACK_START(unsigned char, stack, 8);
    ADD_TO_STACK4(stack, seconds, 0);
    ADD_TO_STACK4(stack, &DateTimeType_68K1, 4);
    STACK_END(stack);
    gCall68KFuncP( gEmulStateP, 
                               PceNativeTrapNo(sysTrapTimSecondsToDateTime),
                               stack, 8 );
    SWAP_DATETIMETYPE_68K_TO_ARM( (char*)dateTimeP, (char*)&DateTimeType_68K1 );
    }
    FUNC_TAIL(TimSecondsToDateTime);
    EMIT_NAME("TimSecondsToDateTime","'T','i','m','S','e','c','o','n','d','s','T','o','D','a','t','e','T','i','m','e'");
}

//Int16 strcasecmp( const Char* s1, const Char* s2 )
//{
//	return StrCaselessCompare( s1, s2 );
//}

Int16
StrCaselessCompare( const Char* s1, const Char* s2 )
{
    Int16 result;
    FUNC_HEADER(StrCaselessCompare);
    {
    STACK_START(unsigned char, stack, 8);
    ADD_TO_STACK4(stack, s1, 0);
    ADD_TO_STACK4(stack, s2, 4);
    STACK_END(stack);
    result = (Int16)gCall68KFuncP( gEmulStateP, 
                               PceNativeTrapNo(sysTrapStrCaselessCompare),
                               stack, 8 );
    }
    FUNC_TAIL(StrCaselessCompare);
    EMIT_NAME("StrCaselessCompare","'S','t','r','C','a','s','e','l','e','s','s','C','o','m','p','a','r','e'");
    return result;
}

Int16 strncasecmp( const Char* s1, const Char* s2 ,int n)
{
	return StrNCaselessCompare( s1, s2 ,n );
}

Int16
StrNCaselessCompare( const Char* s1, const Char* s2 , Int32 n)
{
    Int16 result;
    FUNC_HEADER(StrCaselessCompare);
    {
    STACK_START(unsigned char, stack, 12);
    ADD_TO_STACK4(stack, s1, 0);
    ADD_TO_STACK4(stack, s2, 4);
    ADD_TO_STACK4(stack, n, 8);
    STACK_END(stack);
    result = (Int16)gCall68KFuncP( gEmulStateP, 
                               PceNativeTrapNo(sysTrapStrNCaselessCompare),
                               stack, 12 );
    }
    FUNC_TAIL(StrCaselessCompare);
    EMIT_NAME("StrCaselessCompare","'S','t','r','C','a','s','e','l','e','s','s','C','o','m','p','a','r','e'");
    return result;
}

Err
VFSFileSize( FileRef fileRef, UInt32* fileSizeP )
{
    Err result;
    FUNC_HEADER(VFSFileSize);
    SWAP4_NON_NULL_IN(fileSizeP);
    {
    STACK_START(unsigned char, stack, 8);
    ADD_TO_STACK4(stack, fileRef, 0);
    ADD_TO_STACK4(stack, fileSizeP, 4);
    STACK_END(stack);
    SET_SEL_REG(vfsTrapFileSize);
    result = (Err)gCall68KFuncP( gEmulStateP, 
                               PceNativeTrapNo(sysTrapFileSystemDispatch),
                               stack, 8 );
    SWAP4_NON_NULL_OUT(fileSizeP);
    }
    FUNC_TAIL(VFSFileSize);
    EMIT_NAME("VFSFileSize","'V','F','S','F','i','l','e','S','i','z','e'");
    return result;
}

Err
VFSFileGetAttributes( FileRef fileRef, UInt32* attributesP )
{
    Err result;
    FUNC_HEADER(VFSFileGetAttributes);
    SWAP4_NON_NULL_IN(attributesP);
    {
    STACK_START(unsigned char, stack, 8);
    ADD_TO_STACK4(stack, fileRef, 0);
    ADD_TO_STACK4(stack, attributesP, 4);
    STACK_END(stack);
    SET_SEL_REG(vfsTrapFileGetAttributes);
    result = (Err)gCall68KFuncP( gEmulStateP, 
                               PceNativeTrapNo(sysTrapFileSystemDispatch),
                               stack, 8 );
    SWAP4_NON_NULL_OUT(attributesP);
    }
    FUNC_TAIL(VFSFileGetAttributes);
    EMIT_NAME("VFSFileGetAttributes","'V','F','S','F','i','l','e','G','e','t','A','t','t','r','i','b','u','t','e','s'");
    return result;
}

Err
VFSFileGetDate( FileRef fileRef, UInt16 whichDate, UInt32* dateP )
{
    Err result;
    FUNC_HEADER(VFSFileGetDate);
    SWAP4_NON_NULL_IN(dateP);
    {
    STACK_START(unsigned char, stack, 10);
    ADD_TO_STACK4(stack, fileRef, 0);
    ADD_TO_STACK2(stack, whichDate, 4);
    ADD_TO_STACK4(stack, dateP, 6);
    STACK_END(stack);
    SET_SEL_REG(vfsTrapFileGetDate);
    result = (Err)gCall68KFuncP( gEmulStateP, 
                               PceNativeTrapNo(sysTrapFileSystemDispatch),
                               stack, 10 );
    SWAP4_NON_NULL_OUT(dateP);
    }
    FUNC_TAIL(VFSFileGetDate);
    EMIT_NAME("VFSFileGetDate","'V','F','S','F','i','l','e','G','e','t','D','a','t','e'");
    return result;
}

Err
VFSFileResize( FileRef fileRef, UInt32 newSize )
{
    Err result;
    FUNC_HEADER(VFSFileResize);
    {
    STACK_START(unsigned char, stack, 8);
    ADD_TO_STACK4(stack, fileRef, 0);
    ADD_TO_STACK4(stack, newSize, 4);
    STACK_END(stack);
    SET_SEL_REG(vfsTrapFileResize);
    result = (Err)gCall68KFuncP( gEmulStateP, 
                               PceNativeTrapNo(sysTrapFileSystemDispatch),
                               stack, 8 );
    }
    FUNC_TAIL(VFSFileResize);
    EMIT_NAME("VFSFileResize","'V','F','S','F','i','l','e','R','e','s','i','z','e'");
    return result;
}

Err
VFSDirCreate( UInt16 volRefNum, const Char* dirNameP )
{
    Err result;
    FUNC_HEADER(VFSDirCreate);
    {
    STACK_START(unsigned char, stack, 6);
    ADD_TO_STACK2(stack, volRefNum, 0);
    ADD_TO_STACK4(stack, dirNameP, 2);
    STACK_END(stack);
    SET_SEL_REG(vfsTrapDirCreate);
    result = (Err)gCall68KFuncP( gEmulStateP, 
                               PceNativeTrapNo(sysTrapFileSystemDispatch),
                               stack, 6 );
    }
    FUNC_TAIL(VFSDirCreate);
    EMIT_NAME("VFSDirCreate","'V','F','S','D','i','r','C','r','e','a','t','e'");
    return result;
}

Err
VFSFileDelete( UInt16 volRefNum, const Char* pathNameP )
{
    Err result;
    FUNC_HEADER(VFSFileDelete);
    {
    STACK_START(unsigned char, stack, 6);
    ADD_TO_STACK2(stack, volRefNum, 0);
    ADD_TO_STACK4(stack, pathNameP, 2);
    STACK_END(stack);
    SET_SEL_REG(vfsTrapFileDelete);
    result = (Err)gCall68KFuncP( gEmulStateP, 
                               PceNativeTrapNo(sysTrapFileSystemDispatch),
                               stack, 6 );
    }
    FUNC_TAIL(VFSFileDelete);
    EMIT_NAME("VFSFileDelete","'V','F','S','F','i','l','e','D','e','l','e','t','e'");
    return result;
}

Err
VFSFileRename( UInt16 volRefNum, const Char* pathNameP, const Char* newNameP )
{
    Err result;
    FUNC_HEADER(VFSFileRename);
    {
    STACK_START(unsigned char, stack, 10);
    ADD_TO_STACK2(stack, volRefNum, 0);
    ADD_TO_STACK4(stack, pathNameP, 2);
    ADD_TO_STACK4(stack, newNameP, 6);
    STACK_END(stack);
    SET_SEL_REG(vfsTrapFileRename);
    result = (Err)gCall68KFuncP( gEmulStateP, 
                               PceNativeTrapNo(sysTrapFileSystemDispatch),
                               stack, 10 );
    }
    FUNC_TAIL(VFSFileRename);
    EMIT_NAME("VFSFileRename","'V','F','S','F','i','l','e','R','e','n','a','m','e'");
    return result;
}

Err
VFSFileSetDate( FileRef fileRef, UInt16 whichDate, UInt32 date )
{
    Err result;
    FUNC_HEADER(VFSFileSetDate);
    {
    STACK_START(unsigned char, stack, 10);
    ADD_TO_STACK4(stack, fileRef, 0);
    ADD_TO_STACK2(stack, whichDate, 4);
    ADD_TO_STACK4(stack, date, 6);
    STACK_END(stack);
    SET_SEL_REG(vfsTrapFileSetDate);
    result = (Err)gCall68KFuncP( gEmulStateP, 
                               PceNativeTrapNo(sysTrapFileSystemDispatch),
                               stack, 10 );
    }
    FUNC_TAIL(VFSFileSetDate);
    EMIT_NAME("VFSFileSetDate","'V','F','S','F','i','l','e','S','e','t','D','a','t','e'");
    return result;
}

Err
VFSVolumeSize( UInt16 volRefNum, UInt32* volumeUsedP, UInt32* volumeTotalP )
{
    Err result;
    FUNC_HEADER(VFSVolumeSize);
    SWAP4_NON_NULL_IN(volumeUsedP);
    SWAP4_NON_NULL_IN(volumeTotalP);
    {
    STACK_START(unsigned char, stack, 10);
    ADD_TO_STACK2(stack, volRefNum, 0);
    ADD_TO_STACK4(stack, volumeUsedP, 2);
    ADD_TO_STACK4(stack, volumeTotalP, 6);
    STACK_END(stack);
    SET_SEL_REG(vfsTrapVolumeSize);
    result = (Err)gCall68KFuncP( gEmulStateP, 
                               PceNativeTrapNo(sysTrapFileSystemDispatch),
                               stack, 10 );
    SWAP4_NON_NULL_OUT(volumeUsedP);
    SWAP4_NON_NULL_OUT(volumeTotalP);
    }
    FUNC_TAIL(VFSVolumeSize);
    EMIT_NAME("VFSVolumeSize","'V','F','S','V','o','l','u','m','e','S','i','z','e'");
    return result;
}

Int32
StrAToI( const Char* str )
{
    Int32 result;
    FUNC_HEADER(StrAToI);
    {
    STACK_START(unsigned char, stack, 4);
    ADD_TO_STACK4(stack, str, 0);
    STACK_END(stack);
    result = (Int32)gCall68KFuncP( gEmulStateP, 
                               PceNativeTrapNo(sysTrapStrAToI),
                               stack, 4 );
    }
    FUNC_TAIL(StrAToI);
    EMIT_NAME("StrAToI","'S','t','r','A','T','o','I'");
    return result;
}

Char*
StrNCopy( Char* dst, const Char* src, Int16 n )
{
    Char* result;
    FUNC_HEADER(StrNCopy);
    {
    STACK_START(unsigned char, stack, 10);
    ADD_TO_STACK4(stack, dst, 0);
    ADD_TO_STACK4(stack, src, 4);
    ADD_TO_STACK2(stack, n, 8);
    STACK_END(stack);
    result = (Char*)gCall68KFuncP( gEmulStateP, 
                               PceNativeTrapNo(sysTrapStrNCopy),
                               stack, kPceNativeWantA0 | 10 );
    }
    FUNC_TAIL(StrNCopy);
    EMIT_NAME("StrNCopy","'S','t','r','N','C','o','p','y'");
    return result;
}

Char*
StrStr( const Char* str, const Char* token )
{
    Char* result;
    FUNC_HEADER(StrStr);
    {
    STACK_START(unsigned char, stack, 8);
    ADD_TO_STACK4(stack, str, 0);
    ADD_TO_STACK4(stack, token, 4);
    STACK_END(stack);
    result = (Char*)gCall68KFuncP( gEmulStateP,
                               PceNativeTrapNo(sysTrapStrStr),
                               stack, kPceNativeWantA0 | 8 );
    }
    FUNC_TAIL(StrStr);
    EMIT_NAME("StrStr","'S','t','r','S','t','r'");
    return result;
}

Err
VFSFileWrite( FileRef fileRef, UInt32 numBytes, const void* dataP, UInt32* numBytesWrittenP )
{
    Err result;
    FUNC_HEADER(VFSFileWrite);
    SWAP4_NON_NULL_IN(numBytesWrittenP);
    {
    STACK_START(unsigned char, stack, 16);
    ADD_TO_STACK4(stack, fileRef, 0);
    ADD_TO_STACK4(stack, numBytes, 4);
    ADD_TO_STACK4(stack, dataP, 8);
    ADD_TO_STACK4(stack, numBytesWrittenP, 12);
    STACK_END(stack);
    SET_SEL_REG(vfsTrapFileWrite);
    result = (Err)gCall68KFuncP( gEmulStateP, 
                               PceNativeTrapNo(sysTrapFileSystemDispatch),
                               stack, 16 );
    SWAP4_NON_NULL_OUT(numBytesWrittenP);
    }
    FUNC_TAIL(VFSFileWrite);
    EMIT_NAME("VFSFileWrite","'V','F','S','F','i','l','e','W','r','i','t','e'");
    return result;
}

Char*
StrNCat( Char* dst, const Char* src, Int16 n )
{
    Char* result;
    FUNC_HEADER(StrNCat);
    {
    STACK_START(unsigned char, stack, 10);
    ADD_TO_STACK4(stack, dst, 0);
    ADD_TO_STACK4(stack, src, 4);
    ADD_TO_STACK2(stack, n, 8);
    STACK_END(stack);
    result = (Char*)gCall68KFuncP( gEmulStateP,
                               PceNativeTrapNo(sysTrapStrNCat),
                               stack, kPceNativeWantA0 | 10 );
    }
    FUNC_TAIL(StrNCat);
    EMIT_NAME("StrNCat","'S','t','r','N','C','a','t'");
    return result;
}

Int16
SysRandom( Int32 newSeed )
{
    Int16 result;
    FUNC_HEADER(SysRandom);
    {
    STACK_START(unsigned char, stack, 4);
    ADD_TO_STACK4(stack, newSeed, 0);
    STACK_END(stack);
    result = (Int16)gCall68KFuncP( gEmulStateP,
                               PceNativeTrapNo(sysTrapSysRandom),
                               stack, 4 );
    }
    FUNC_TAIL(SysRandom);
    EMIT_NAME("SysRandom","'S','y','s','R','a','n','d','o','m'");
    return result;
}

Err
MemChunkFree( MemPtr chunkDataP )
{
    Err result;
    FUNC_HEADER(MemChunkFree);
    {
    STACK_START(unsigned char, stack, 4);
    ADD_TO_STACK4(stack, chunkDataP, 0);
    STACK_END(stack);
    result = (Err)gCall68KFuncP( gEmulStateP, 
                               PceNativeTrapNo(sysTrapMemChunkFree),
                               stack, 4 );
    }
    FUNC_TAIL(MemChunkFree);
    EMIT_NAME("MemChunkFree","'M','e','m','C','h','u','n','k','F','r','e','e'");
    return result;
}

MemPtr
MemPtrNew( UInt32 size )
{
    MemPtr result;
    FUNC_HEADER(MemPtrNew);
    {
    STACK_START(unsigned char, stack, 4);
    ADD_TO_STACK4(stack, size, 0);
    STACK_END(stack);
    result = (MemPtr)gCall68KFuncP( gEmulStateP, 
                               PceNativeTrapNo(sysTrapMemPtrNew),
                               stack, kPceNativeWantA0 | 4 );
    }
    FUNC_TAIL(MemPtrNew);
    EMIT_NAME("MemPtrNew","'M','e','m','P','t','r','N','e','w'");
    return result;
}

BitmapType*
BmpCreate( Coord width, Coord height, UInt8 depth, ColorTableType* colortableP, UInt16* error )
{
    BitmapType* result;
    FUNC_HEADER(BmpCreate);
    SWAP2_NON_NULL_IN(error);
    {
    STACK_START(unsigned char, stack, 14);
    ADD_TO_STACK2(stack, width, 0);
    ADD_TO_STACK2(stack, height, 2);
    ADD_TO_STACK1(stack, depth, 4);
    ADD_TO_STACK4(stack, colortableP, 6);
    ADD_TO_STACK4(stack, error, 10);
    STACK_END(stack);
    result = (BitmapType*)gCall68KFuncP( gEmulStateP, 
                               PceNativeTrapNo(sysTrapBmpCreate),
                               stack, kPceNativeWantA0 | 14 );
    SWAP2_NON_NULL_OUT(error);
    }
    FUNC_TAIL(BmpCreate);
    EMIT_NAME("BmpCreate","'B','m','p','C','r','e','a','t','e'");
    return result;
}

BitmapTypeV3*
BmpCreateBitmapV3( const BitmapType* bitmapP, UInt16 density, const void* bitsP, const ColorTableType* colorTableP )
{
    BitmapTypeV3* result;
    FUNC_HEADER(BmpCreateBitmapV3);
    {
    STACK_START(unsigned char, stack, 14);
    ADD_TO_STACK4(stack, bitmapP, 0);
    ADD_TO_STACK2(stack, density, 4);
    ADD_TO_STACK4(stack, bitsP, 6);
    ADD_TO_STACK4(stack, colorTableP, 10);
    STACK_END(stack);
    SET_SEL_REG(HDSelectorBmpCreateBitmapV3);
    result = (BitmapTypeV3*)gCall68KFuncP( gEmulStateP, 
                               PceNativeTrapNo(sysTrapHighDensityDispatch),
                               stack, kPceNativeWantA0 | 14 );
    }
    FUNC_TAIL(BmpCreateBitmapV3);
    EMIT_NAME("BmpCreateBitmapV3","'B','m','p','C','r','e','a','t','e','B','i','t','m','a','p','V','3'");
    return result;
}

void
WinDrawBitmap( BitmapPtr bitmapP, Coord x, Coord y )
{
    FUNC_HEADER(WinDrawBitmap);
    {
    STACK_START(unsigned char, stack, 8);
    ADD_TO_STACK4(stack, bitmapP, 0);
    ADD_TO_STACK2(stack, x, 4);
    ADD_TO_STACK2(stack, y, 6);
    STACK_END(stack);
    gCall68KFuncP( gEmulStateP, 
                               PceNativeTrapNo(sysTrapWinDrawBitmap),
                               stack, 8 );
    }
    FUNC_TAIL(WinDrawBitmap);
    EMIT_NAME("WinDrawBitmap","'W','i','n','D','r','a','w','B','i','t','m','a','p'");
}

Err
BmpDelete( BitmapType* bitmapP )
{
    Err result;
    FUNC_HEADER(BmpDelete);
    {
    STACK_START(unsigned char, stack, 4);
    ADD_TO_STACK4(stack, bitmapP, 0);
    STACK_END(stack);
    result = (Err)gCall68KFuncP( gEmulStateP, 
                               PceNativeTrapNo(sysTrapBmpDelete),
                               stack, 4 );
    }
    FUNC_TAIL(BmpDelete);
    EMIT_NAME("BmpDelete","'B','m','p','D','e','l','e','t','e'");
    return result;
}

/*
void SWAP_EVENTTYPE (char *src){
	char a;
	a=*(src+1);
	*(src+1)=*src;
	*src=a;
	a=*(src+4);
	*(src+4)=*(src+5);
	*(src+5)=a;
	a=*(src+6);
	*(src+6)=*(src+7);
	*(src+7)=a;

	*dest=*(src+1);
	*(dest+1)=*src;
	*(dest+2)=*(src+2);
	*(dest+3)=*(src+3);
	*(dest+4)=*(src+5);
	*(dest+5)=*(src+4);
	*(dest+6)=*(src+7);
	*(dest+7)=*(src+6);
	if ((*p==keyDownEvent)||(*p==keyUpEvent)||(*p==keyHoldEvent)){
		*(dest+8)=*(src+9);
		*(dest+9)=*(src+8);
	}
}
*/
//extern char warning_buffer[];

void
EvtGetEvent( EventType* event, Int32 timeout )
{
    FUNC_HEADER(EvtGetEvent);
//    EventType EventType_68K0;
//    SWAP_EVENTTYPE( (char *)&EventType_68K0, (char *)event );
    {
    STACK_START(unsigned char, stack, 8);
    ADD_TO_STACK4(stack, event, 0);
    ADD_TO_STACK4(stack, timeout, 4);
    STACK_END(stack);

    gCall68KFuncP( gEmulStateP, 
                               PceNativeTrapNo(sysTrapEvtGetEvent),
                               stack, 8 );
//    SWAP_EVENTTYPE( (char *)event );
    }
    FUNC_TAIL(EvtGetEvent);
    EMIT_NAME("EvtGetEvent","'E','v','t','G','e','t','E','v','e','n','t'");
}

Err
VFSVolumeEnumerate( UInt16* volRefNumP, UInt32* volIteratorP )
{
    Err result;
    FUNC_HEADER(VFSVolumeEnumerate);
    SWAP2_NON_NULL_IN(volRefNumP);
    SWAP4_NON_NULL_IN(volIteratorP);
    {
    STACK_START(unsigned char, stack, 8);
    ADD_TO_STACK4(stack, volRefNumP, 0);
    ADD_TO_STACK4(stack, volIteratorP, 4);
    STACK_END(stack);
    SET_SEL_REG(vfsTrapVolumeEnumerate);
    result = (Err)gCall68KFuncP( gEmulStateP,
                               PceNativeTrapNo(sysTrapFileSystemDispatch),
                               stack, 8 );
    SWAP2_NON_NULL_OUT(volRefNumP);
    SWAP4_NON_NULL_OUT(volIteratorP);
    }
    FUNC_TAIL(VFSVolumeEnumerate);
    EMIT_NAME("VFSVolumeEnumerate","'V','F','S','V','o','l','u','m','e','E','n','u','m','e','r','a','t','e'");
    return result;
}

Err
VFSFileSetAttributes( FileRef fileRef, UInt32 attributes )
{
    Err result;
    FUNC_HEADER(VFSFileSetAttributes);
    {
    STACK_START(unsigned char, stack, 8);
    ADD_TO_STACK4(stack, fileRef, 0);
    ADD_TO_STACK4(stack, attributes, 4);
    STACK_END(stack);
    SET_SEL_REG(vfsTrapFileSetAttributes);
    result = (Err)gCall68KFuncP( gEmulStateP,
                               PceNativeTrapNo(sysTrapFileSystemDispatch),
                               stack, 8 );
    }
    FUNC_TAIL(VFSFileSetAttributes);
    EMIT_NAME("VFSFileSetAttributes","'V','F','S','F','i','l','e','S','e','t','A','t','t','r','i','b','u','t','e','s'");
    return result;
}

void
WinDrawChars( const Char* chars, Int16 len, Coord x, Coord y )
{
    FUNC_HEADER(WinDrawChars);
    {
    STACK_START(unsigned char, stack, 10);
    ADD_TO_STACK4(stack, chars, 0);
    ADD_TO_STACK2(stack, len, 4);
    ADD_TO_STACK2(stack, x, 6);
    ADD_TO_STACK2(stack, y, 8);
    STACK_END(stack);
    gCall68KFuncP( gEmulStateP,
                               PceNativeTrapNo(sysTrapWinDrawChars),
                               stack, 10 );
    }
    FUNC_TAIL(WinDrawChars);
    EMIT_NAME("WinDrawChars","'W','i','n','D','r','a','w','C','h','a','r','s'");
}

void*
BmpGetBits( BitmapType* bitmapP )
{
    void* result;
    FUNC_HEADER(BmpGetBits);
    {
    STACK_START(unsigned char, stack, 4);
    ADD_TO_STACK4(stack, bitmapP, 0);
    STACK_END(stack);
    result = (void*)gCall68KFuncP( gEmulStateP,
                               PceNativeTrapNo(sysTrapBmpGetBits),
                               stack, kPceNativeWantA0 | 4 );
    }
    FUNC_TAIL(BmpGetBits);
    EMIT_NAME("BmpGetBits","'B','m','p','G','e','t','B','i','t','s'");
    return result;
}

Int16
StrPrintF( Char* s, const Char* formatStr, ... )
{
    Int16 result;
    UInt32 x1, x2, x3, x4, x5, x6, x7, x8;
    x1=(UInt32)&formatStr;
    x2=x1-4;
    x3=x2-4;
    x4=x3-4;
    x5=x4-4;
    x6=x5-4;
    x7=x6-4;
    x8=x7-4;
    FUNC_HEADER(StrPrintF);
    {
    STACK_START(unsigned char, stack, 40);
    ADD_TO_STACK4(stack, s, 0);
    ADD_TO_STACK4(stack, formatStr, 4);
    ADD_TO_STACK4(stack, x1, 8);
    ADD_TO_STACK4(stack, x2, 12);
    ADD_TO_STACK4(stack, x3, 16);
    ADD_TO_STACK4(stack, x4, 20);
    ADD_TO_STACK4(stack, x5, 24);
    ADD_TO_STACK4(stack, x6, 28);
    ADD_TO_STACK4(stack, x7, 32);
    ADD_TO_STACK4(stack, x8, 36);
    STACK_END(stack);
    result = (Int16)gCall68KFuncP( gEmulStateP,
                               PceNativeTrapNo(sysTrapStrPrintF),
                               stack, 40 );
    }
    FUNC_TAIL(StrPrintF);
    EMIT_NAME("StrPrintF","'S','t','r','P','r','i','n','t','F'");
    return result;
}
/*
int vsprintf (char *a, const char *b, va_list c){
	return (int)(StrVPrintF(a,b,(_Palm_va_list)c));
}

Int16
StrVPrintF( Char* s, const Char* format, _Palm_va_list arg )
{
    Int16 result;
    FUNC_HEADER(StrVPrintF);
    {
    STACK_START(unsigned char, stack, 12);
    ADD_TO_STACK4(stack, s, 0);
    ADD_TO_STACK4(stack, format, 4);
    ADD_TO_STACK4(stack, arg, 8);
    STACK_END(stack);
    result = (Int16)gCall68KFuncP( gEmulStateP,
                               PceNativeTrapNo(sysTrapStrVPrintF),
                               stack, 12 );
    }
    FUNC_TAIL(StrVPrintF);
    EMIT_NAME("StrVPrintF","'S','t','r','V','P','r','i','n','t','F'");
    return result;
}
*/
Int16
aStrPrintF( Char* s, const Char* formatStr, ... )
{
    Int16 result;
    int x1, x2, x3, x4, x5, x6, x7, x8;
    va_list ap;
    va_start (ap, formatStr);
    {
	x1 = va_arg (ap, int);
	x2 = va_arg (ap, int);
	x3 = va_arg (ap, int);
	x4 = va_arg (ap, int);
	x5 = va_arg (ap, int);
	x6 = va_arg (ap, int);
	x7 = va_arg (ap, int);
	x8 = va_arg (ap, int);
    }
    FUNC_HEADER(StrPrintF);
    {
    STACK_START(unsigned char, stack, 40);
    ADD_TO_STACK4(stack, s, 0);
    ADD_TO_STACK4(stack, formatStr, 4);
    ADD_TO_STACK4(stack, x1, 8);
    ADD_TO_STACK4(stack, x2, 12);
    ADD_TO_STACK4(stack, x3, 16);
    ADD_TO_STACK4(stack, x4, 20);
    ADD_TO_STACK4(stack, x5, 24);
    ADD_TO_STACK4(stack, x6, 28);
    ADD_TO_STACK4(stack, x7, 32);
    ADD_TO_STACK4(stack, x8, 36);
    STACK_END(stack);
    result = (Int16)gCall68KFuncP( gEmulStateP,
                               PceNativeTrapNo(sysTrapStrPrintF),
                               stack, 40 );
    }
    FUNC_TAIL(StrPrintF);
    EMIT_NAME("StrPrintF","'S','t','r','P','r','i','n','t','F'");
    return result;
}

double FlpATof( const Char* s )
{
    double result;
    Int32 a;
    FUNC_HEADER(FlpAToF);
    {
    STACK_START(unsigned char, stack, 4);
    ADD_TO_STACK4(stack, s, 0);
    STACK_END(stack);
    SET_SEL_REG(sysFloatAToF);
    result = (double)gCall68KFuncP( gEmulStateP,
                               PceNativeTrapNo(sysTrapFlpDispatch),
                               stack, 4 );
    }
    FUNC_TAIL(FlpAToF);
    EMIT_NAME("FlpAToF","'F','l','p','A','T','o','F'");
    return result;
}

MemPtr
MemChunkNew( UInt16 heapID, UInt32 size, UInt16 attr )
{
    MemPtr result;
    FUNC_HEADER(MemChunkNew);
    {
    STACK_START(unsigned char, stack, 8);
    ADD_TO_STACK2(stack, heapID, 0);
    ADD_TO_STACK4(stack, size, 2);
    ADD_TO_STACK2(stack, attr, 6);
    STACK_END(stack);
    result = (MemPtr)gCall68KFuncP( gEmulStateP,
                               PceNativeTrapNo(sysTrapMemChunkNew),
                               stack, kPceNativeWantA0 | 8 );
    }
    FUNC_TAIL(MemChunkNew);
    EMIT_NAME("MemChunkNew","'M','e','m','C','h','u','n','k','N','e','w'");
    return result;
}

Char*
StrIToA( Char* s, Int32 i )
{
    Char* result;
    FUNC_HEADER(StrIToA);
    {
    STACK_START(unsigned char, stack, 8);
    ADD_TO_STACK4(stack, s, 0);
    ADD_TO_STACK4(stack, i, 4);
    STACK_END(stack);
    result = (Char*)gCall68KFuncP( gEmulStateP,
                               PceNativeTrapNo(sysTrapStrIToA),
                               stack, kPceNativeWantA0 | 8 );
    }
    FUNC_TAIL(StrIToA);
    EMIT_NAME("StrIToA","'S','t','r','I','T','o','A'");
    return result;
}

Char*
StrIToH( Char* s, UInt32 i )
{
    Char* result;
    FUNC_HEADER(StrIToA);
    {
    STACK_START(unsigned char, stack, 8);
    ADD_TO_STACK4(stack, s, 0);
    ADD_TO_STACK4(stack, i, 4);
    STACK_END(stack);
    result = (Char*)gCall68KFuncP( gEmulStateP,
                               PceNativeTrapNo(sysTrapStrIToH),
                               stack, kPceNativeWantA0 | 8 );
    }
    FUNC_TAIL(StrIToA);
    EMIT_NAME("StrIToA","'S','t','r','I','T','o','A'");
    return result;
}

Err
MemHeapFreeBytes( UInt16 heapID, UInt32* freeP, UInt32* maxP )
{
    Err result;
    FUNC_HEADER(MemHeapFreeBytes);
    SWAP4_NON_NULL_IN(freeP);
    SWAP4_NON_NULL_IN(maxP);
    {
    STACK_START(unsigned char, stack, 10);
    ADD_TO_STACK2(stack, heapID, 0);
    ADD_TO_STACK4(stack, freeP, 2);
    ADD_TO_STACK4(stack, maxP, 6);
    STACK_END(stack);
    result = (Err)gCall68KFuncP( gEmulStateP,
                               PceNativeTrapNo(sysTrapMemHeapFreeBytes),
                               stack, 10 );
    SWAP4_NON_NULL_OUT(freeP);
    SWAP4_NON_NULL_OUT(maxP);
    }
    FUNC_TAIL(MemHeapFreeBytes);
    EMIT_NAME("MemHeapFreeBytes","'M','e','m','H','e','a','p','F','r','e','e','B','y','t','e','s'");
    return result;
}

WinHandle
WinCreateBitmapWindow( BitmapType* bitmapP, UInt16* error )
{
    WinHandle result;
    FUNC_HEADER(WinCreateBitmapWindow);
    SWAP2_NON_NULL_IN(error);
    {
    STACK_START(unsigned char, stack, 8);
    ADD_TO_STACK4(stack, bitmapP, 0);
    ADD_TO_STACK4(stack, error, 4);
    STACK_END(stack);
    result = (WinHandle)gCall68KFuncP( gEmulStateP,
                               PceNativeTrapNo(sysTrapWinCreateBitmapWindow),
                               stack, kPceNativeWantA0 | 8 );
    SWAP2_NON_NULL_OUT(error);
    }
    FUNC_TAIL(WinCreateBitmapWindow);
    EMIT_NAME("WinCreateBitmapWindow","'W','i','n','C','r','e','a','t','e','B','i','t','m','a','p','W','i','n','d','o','w'");
    return result;
}

void
WinSetActiveWindow( WinHandle winHandle )
{
    FUNC_HEADER(WinSetActiveWindow);
    {
    STACK_START(unsigned char, stack, 4);
    ADD_TO_STACK4(stack, winHandle, 0);
    STACK_END(stack);
    gCall68KFuncP( gEmulStateP,
                               PceNativeTrapNo(sysTrapWinSetActiveWindow),
                               stack, 4 );
    }
    FUNC_TAIL(WinSetActiveWindow);
    EMIT_NAME("WinSetActiveWindow","'W','i','n','S','e','t','A','c','t','i','v','e','W','i','n','d','o','w'");
}

UInt32
MemPtrSize( MemPtr p )
{
    UInt32 result;
    FUNC_HEADER(MemPtrSize);
    {
    STACK_START(unsigned char, stack, 4);
    ADD_TO_STACK4(stack, p, 0);
    STACK_END(stack);
    result = (UInt32)gCall68KFuncP( gEmulStateP,
                               PceNativeTrapNo(sysTrapMemPtrSize),
                               stack, 4 );
    }
    FUNC_TAIL(MemPtrSize);
    EMIT_NAME("MemPtrSize","'M','e','m','P','t','r','S','i','z','e'");
    return result;
}


Err
WinPalette( UInt8 operation, Int16 startIndex, UInt16 paletteEntries, RGBColorType* tableP )
{
    Err result;
    FUNC_HEADER(WinPalette);
    {
    STACK_START(unsigned char, stack, 10);
    ADD_TO_STACK1(stack, operation, 0);
    ADD_TO_STACK2(stack, startIndex, 2);
    ADD_TO_STACK2(stack, paletteEntries, 4);
    ADD_TO_STACK4(stack, tableP, 6);
    STACK_END(stack);
    result = (Err)gCall68KFuncP( gEmulStateP,
                               PceNativeTrapNo(sysTrapWinPalette),
                               stack, 10 );
    }
    FUNC_TAIL(WinPalette);
    EMIT_NAME("WinPalette","'W','i','n','P','a','l','e','t','t','e'");
    return result;
}

Err
WinScreenMode( WinScreenModeOperation operation, UInt32* widthP, UInt32* heightP, UInt32* depthP, Boolean* enableColorP )
{
    Err result;
    FUNC_HEADER(WinScreenMode);
    SWAP4_NON_NULL_IN(widthP);
    SWAP4_NON_NULL_IN(heightP);
    SWAP4_NON_NULL_IN(depthP);
    SWAP1_NON_NULL_IN(enableColorP);
    {
    STACK_START(unsigned char, stack, 18);
    ADD_TO_STACK1(stack, operation, 0);
    ADD_TO_STACK4(stack, widthP, 2);
    ADD_TO_STACK4(stack, heightP, 6);
    ADD_TO_STACK4(stack, depthP, 10);
    ADD_TO_STACK4(stack, enableColorP, 14);
    STACK_END(stack);
    result = (Err)gCall68KFuncP( gEmulStateP,
                               PceNativeTrapNo(sysTrapWinScreenMode),
                               stack, 18 );
    SWAP4_NON_NULL_OUT(widthP);
    SWAP4_NON_NULL_OUT(heightP);
    SWAP4_NON_NULL_OUT(depthP);
    SWAP1_NON_NULL_OUT(enableColorP);
    }
    FUNC_TAIL(WinScreenMode);
    EMIT_NAME("WinScreenMode","'W','i','n','S','c','r','e','e','n','M','o','d','e'");
    return result;
}


UInt32
KeyCurrentState(  )
{
    UInt32 result;
    FUNC_HEADER(KeyCurrentState);
    {
    STACK_START(unsigned char, stack, 0);
    STACK_END(stack);
    result = (UInt32)gCall68KFuncP( gEmulStateP,
                               PceNativeTrapNo(sysTrapKeyCurrentState),
                               stack, 0 );
    }
    FUNC_TAIL(KeyCurrentState);
    EMIT_NAME("KeyCurrentState","'K','e','y','C','u','r','r','e','n','t','S','t','a','t','e'");
    return result;
}

Err
DmWrite( void* recordP, UInt32 offset, const void* srcP, UInt32 bytes )
{
    Err result;
    FUNC_HEADER(DmWrite);
    {
    STACK_START(unsigned char, stack, 16);
    ADD_TO_STACK4(stack, recordP, 0);
    ADD_TO_STACK4(stack, offset, 4);
    ADD_TO_STACK4(stack, srcP, 8);
    ADD_TO_STACK4(stack, bytes, 12);
    STACK_END(stack);
    result = (Err)gCall68KFuncP( gEmulStateP,
                               PceNativeTrapNo(sysTrapDmWrite),
                               stack, 16 );
    }
    FUNC_TAIL(DmWrite);
    EMIT_NAME("DmWrite","'D','m','W','r','i','t','e'");
    return result;
}

WinHandle
WinGetDisplayWindow(  )
{
    WinHandle result;
    FUNC_HEADER(WinGetDisplayWindow);
    {
    STACK_START(unsigned char, stack, 0);
    STACK_END(stack);
    result = (WinHandle)gCall68KFuncP( gEmulStateP,
                               PceNativeTrapNo(sysTrapWinGetDisplayWindow),
                               stack, kPceNativeWantA0 | 0 );
    }
    FUNC_TAIL(WinGetDisplayWindow);
    EMIT_NAME("WinGetDisplayWindow","'W','i','n','G','e','t','D','i','s','p','l','a','y','W','i','n','d','o','w'");
    return result;
}

BitmapType*
WinGetBitmap( WinHandle winHandle )
{
    BitmapType* result;
    FUNC_HEADER(WinGetBitmap);
    {
    STACK_START(unsigned char, stack, 4);
    ADD_TO_STACK4(stack, winHandle, 0);
    STACK_END(stack);
    result = (BitmapType*)gCall68KFuncP( gEmulStateP,
                               PceNativeTrapNo(sysTrapWinGetBitmap),
                               stack, 4 );
    }
    FUNC_TAIL(WinGetBitmap);
    EMIT_NAME("WinGetBitmap","'W','i','n','G','e','t','B','i','t','m','a','p'");
    return result;
}

Boolean
SysHandleEvent( EventPtr eventP )
{
    Boolean result;
    FUNC_HEADER(SysHandleEvent);
    {
    STACK_START(unsigned char, stack, 4);
    ADD_TO_STACK4(stack, eventP, 0);
    STACK_END(stack);
    result = (Boolean)gCall68KFuncP( gEmulStateP,
                               PceNativeTrapNo(sysTrapSysHandleEvent),
                               stack, 4 );
    }
    FUNC_TAIL(SysHandleEvent);
    EMIT_NAME("SysHandleEvent","'S','y','s','H','a','n','d','l','e','E','v','e','n','t'");
    return result;
}

SysAppInfoPtr
SysGetAppInfo( SysAppInfoPtr* appInfoP1, SysAppInfoPtr* appInfoP2 )
{
    SysAppInfoPtr result;
    FUNC_HEADER(SysHandleEvent);
    {
    STACK_START(unsigned char, stack, 8);
    ADD_TO_STACK4(stack, appInfoP1, 0);
    ADD_TO_STACK4(stack, appInfoP2, 4);
    STACK_END(stack);
    result = (SysAppInfoPtr)gCall68KFuncP( gEmulStateP,
                               PceNativeTrapNo(sysTrapSysGetAppInfo),
                               stack, 8 );
    }
    FUNC_TAIL(SysHandleEvent);
    EMIT_NAME("SysHandleEvent","'S','y','s','H','a','n','d','l','e','E','v','e','n','t'");
    return result;
}


Err
SndStreamCreate( SndStreamRef* channel, SndStreamMode mode, UInt32 samplerate, SndSampleType type, SndStreamWidth width, SndStreamBufferCallback func, void* userdata, UInt32 buffsize, Boolean armNative )
{
    Err result;
    FUNC_HEADER(SndStreamCreate);
    {
    STACK_START(unsigned char, stack, 28);
    ADD_TO_STACK4(stack, channel, 0);
    ADD_TO_STACK1(stack, mode, 4);
    ADD_TO_STACK4(stack, samplerate, 6);
    ADD_TO_STACK2(stack, type, 10);
    ADD_TO_STACK1(stack, width, 12);
    ADD_TO_STACK4(stack, func, 14);
    ADD_TO_STACK4(stack, userdata, 18);
    ADD_TO_STACK4(stack, buffsize, 22);
    ADD_TO_STACK1(stack, armNative, 26);
    STACK_END(stack);
    result = (Err)gCall68KFuncP( gEmulStateP,
                               PceNativeTrapNo(sysTrapSndStreamCreate),
                               stack, 28 );
    }
    *channel=Byte_Swap32(*channel);
    FUNC_TAIL(SndStreamCreate);
    EMIT_NAME("SndStreamCreate","'S','n','d','S','t','r','e','a','m','C','r','e','a','t','e'");
    return result;
} 


Err
SndStreamCreateExtended( SndStreamRef* channel, SndStreamMode mode, SndFormatType format, UInt32 samplerate, SndSampleType type, SndStreamWidth width, SndStreamVariableBufferCallback func, void* userdata, UInt32 buffsize, Boolean armNative )
{
    Err result;
    FUNC_HEADER(SndStreamCreateExtended);
    {
    STACK_START(unsigned char, stack, 32);
    ADD_TO_STACK4(stack, channel, 0);
    ADD_TO_STACK1(stack, mode, 4);
    ADD_TO_STACK4(stack, format, 6);
    ADD_TO_STACK4(stack, samplerate, 10);
    ADD_TO_STACK2(stack, type, 14);
    ADD_TO_STACK1(stack, width, 16);
    ADD_TO_STACK4(stack, func, 18);
    ADD_TO_STACK4(stack, userdata, 22);
    ADD_TO_STACK4(stack, buffsize, 26);
    ADD_TO_STACK1(stack, armNative, 30);
    STACK_END(stack);
    result = (Err)gCall68KFuncP( gEmulStateP,
                               PceNativeTrapNo(sysTrapSndStreamCreateExtended),
                               stack, 32 );
    *channel=Byte_Swap32(*channel);
    }
    FUNC_TAIL(SndStreamCreateExtended);
    EMIT_NAME("SndStreamCreateExtended","'S','n','d','S','t','r','e','a','m','C','r','e','a','t','e','E','x','t','e','n','d','e','d'");
    return result;
}

Err
SndStreamDelete( SndStreamRef channel )
{
    Err result;
    FUNC_HEADER(SndStreamDelete);
    {
    STACK_START(unsigned char, stack, 4);
    ADD_TO_STACK4(stack, channel, 0);
    STACK_END(stack);
    result = (Err)gCall68KFuncP( gEmulStateP,
                               PceNativeTrapNo(sysTrapSndStreamDelete),
                               stack, 4 );
    }
    FUNC_TAIL(SndStreamDelete);
    EMIT_NAME("SndStreamDelete","'S','n','d','S','t','r','e','a','m','D','e','l','e','t','e'");
    return result;
}

Err
SndStreamStart( SndStreamRef channel )
{
    Err result;
    FUNC_HEADER(SndStreamStart);
    {
    STACK_START(unsigned char, stack, 4);
    ADD_TO_STACK4(stack, channel, 0);
    STACK_END(stack);
    result = (Err)gCall68KFuncP( gEmulStateP,
                               PceNativeTrapNo(sysTrapSndStreamStart),
                               stack, 4 );
    }
    FUNC_TAIL(SndStreamStart);
    EMIT_NAME("SndStreamStart","'S','n','d','S','t','r','e','a','m','S','t','a','r','t'");
    return result;
}

Err
SndStreamStop( SndStreamRef channel )
{
    Err result;
    FUNC_HEADER(SndStreamStop);
    {
    STACK_START(unsigned char, stack, 4);
    ADD_TO_STACK4(stack, channel, 0);
    STACK_END(stack);
    result = (Err)gCall68KFuncP( gEmulStateP,
                               PceNativeTrapNo(sysTrapSndStreamStop),
                               stack, 4 );
    }
    FUNC_TAIL(SndStreamStop);
    EMIT_NAME("SndStreamStop","'S','n','d','S','t','r','e','a','m','S','t','o','p'");
    return result;
}

Err
SndStreamPause( SndStreamRef channel, Boolean pause )
{
    Err result;
    FUNC_HEADER(SndStreamPause);
    {
    STACK_START(unsigned char, stack, 6);
    ADD_TO_STACK4(stack, channel, 0);
    ADD_TO_STACK1(stack, pause, 4);
    STACK_END(stack);
    result = (Err)gCall68KFuncP( gEmulStateP,
                               PceNativeTrapNo(sysTrapSndStreamPause),
                               stack, 6 );
    }
    FUNC_TAIL(SndStreamPause);
    EMIT_NAME("SndStreamPause","'S','n','d','S','t','r','e','a','m','P','a','u','s','e'");
    return result;
}

Err
SndStreamGetVolume( SndStreamRef channel, Int32* volume )
{
    Err result;
    FUNC_HEADER(SndStreamGetVolume);
    SWAP4_NON_NULL_IN(volume);
    {
    STACK_START(unsigned char, stack, 8);
    ADD_TO_STACK4(stack, channel, 0);
    ADD_TO_STACK4(stack, volume, 4);
    STACK_END(stack);
    result = (Err)gCall68KFuncP( gEmulStateP,
                               PceNativeTrapNo(sysTrapSndStreamGetVolume),
                               stack, 8 );
    SWAP4_NON_NULL_OUT(volume);
    }
    FUNC_TAIL(SndStreamGetVolume);
    EMIT_NAME("SndStreamGetVolume","'S','n','d','S','t','r','e','a','m','G','e','t','V','o','l','u','m','e'");
    return result;
}

Err
SndStreamSetVolume( SndStreamRef channel, Int32 volume )
{
    Err result;
    FUNC_HEADER(SndStreamSetVolume);
    {
    STACK_START(unsigned char, stack, 8);
    ADD_TO_STACK4(stack, channel, 0);
    ADD_TO_STACK4(stack, volume, 4);
    STACK_END(stack);
    result = (Err)gCall68KFuncP( gEmulStateP,
                               PceNativeTrapNo(sysTrapSndStreamSetVolume),
                               stack, 8 );
    }
    FUNC_TAIL(SndStreamSetVolume);
    EMIT_NAME("SndStreamSetVolume","'S','n','d','S','t','r','e','a','m','S','e','t','V','o','l','u','m','e'");
    return result;
}

Err
EvtResetAutoOffTimer(  )
{
    Err result;
    FUNC_HEADER(EvtResetAutoOffTimer);
    {
    STACK_START(unsigned char, stack, 0);
    STACK_END(stack);
    result = (Err)gCall68KFuncP( gEmulStateP,
                               PceNativeTrapNo(sysTrapEvtResetAutoOffTimer),
                               stack, 0 );
    }
    FUNC_TAIL(EvtResetAutoOffTimer);
    EMIT_NAME("EvtResetAutoOffTimer","'E','v','t','R','e','s','e','t','A','u','t','o','O','f','f','T','i','m','e','r'");
    return result;
}

Err
MemSemaphoreReserve( Boolean writeAccess )
{
    Err result;
    FUNC_HEADER(MemSemaphoreReserve);
    {
    STACK_START(unsigned char, stack, 2);
    ADD_TO_STACK1(stack, writeAccess, 0);
    STACK_END(stack);
    result = (Err)gCall68KFuncP( gEmulStateP,
                               PceNativeTrapNo(sysTrapMemSemaphoreReserve),
                               stack, 2 );
    }
    FUNC_TAIL(MemSemaphoreReserve);
    EMIT_NAME("MemSemaphoreReserve","'M','e','m','S','e','m','a','p','h','o','r','e','R','e','s','e','r','v','e'");
    return result;
}

Err
MemSemaphoreRelease( Boolean writeAccess )
{
    Err result;
    FUNC_HEADER(MemSemaphoreRelease);
    {
    STACK_START(unsigned char, stack, 2);
    ADD_TO_STACK1(stack, writeAccess, 0);
    STACK_END(stack);
    result = (Err)gCall68KFuncP( gEmulStateP,
                               PceNativeTrapNo(sysTrapMemSemaphoreRelease),
                               stack, 2 );
    }
    FUNC_TAIL(MemSemaphoreRelease);
    EMIT_NAME("MemSemaphoreRelease","'M','e','m','S','e','m','a','p','h','o','r','e','R','e','l','e','a','s','e'");
    return result;
}
/*
void HwrEnableDataWrites()
{
     gCall68KFuncP( gEmulStateP,
                               PceNativeTrapNo(sysTrapHwrEnableDataWrites),
                               0, 0 );
}

void HwrDisableDataWrites()
{
     gCall68KFuncP( gEmulStateP,
                               PceNativeTrapNo(sysTrapHwrDisableDataWrites),
                               0, 0 );
}
*/
/*
Err
PA1L_adpcmOpen( UInt16 refNum, UInt8 fs, UInt8 *data, UInt32 dwSize,
		void *callbackInfoP,
		UInt8 *hd, Boolean *retval )
{
    Err result;
    FUNC_HEADER(SndStreamCreate);
    {
    STACK_START(unsigned char, stack, 24);
    ADD_TO_STACK2(stack, refNum, 0);
    ADD_TO_STACK1(stack, fs, 2);
    ADD_TO_STACK4(stack, data, 4);
    ADD_TO_STACK4(stack, dwSize, 8);
    ADD_TO_STACK4(stack, callbackInfoP, 12);
    ADD_TO_STACK4(stack, hd, 16);
    ADD_TO_STACK4(stack, retval, 20);
    STACK_END(stack);
    result = (Err)gCall68KFuncP( gEmulStateP,
                               PceNativeTrapNo(sysLibTrapCustom + 39),
                               stack, 24 );
    }
    hd=Byte_Swap32(hd);
    *hd=Byte_Swap16(*hd);
    FUNC_TAIL(SndStreamCreate);
    EMIT_NAME("SndStreamCreate","'S','n','d','S','t','r','e','a','m','C','r','e','a','t','e'");
    return result;
} 
*/

extern UInt16 volumeRef;

int stat (char* name, struct stat *statbuf){
	FileRef fr;
	int res=0;
	if(VFSFileOpen (volumeRef,name,vfsModeRead,&fr)!=errNone) return -1;
	if(!res)if(VFSFileGetAttributes(fr,&statbuf->st_mode)!=errNone) res= -1;
	if(!res)if(VFSFileGetDate (fr,vfsFileDateModified,&statbuf->date)!=errNone) res= -1;
	if((!res)&&(statbuf->st_mode!=vfsFileAttrDirectory))if(VFSFileSize (fr,&statbuf->filesize)!=errNone) res= -1;
	statbuf->st_size=statbuf->filesize;
	VFSFileClose (fr);
	return res;
}

int fstat (FileRef fr, struct stat *statbuf){
	if(VFSFileGetDate (fr,vfsFileDateModified,&statbuf->date)!=errNone) return -1;
	if(VFSFileSize (fr,&statbuf->filesize)!=errNone) return -1;
	if(VFSFileGetAttributes(fr,&statbuf->st_mode)!=errNone) return -1;
	return 0;
}

typedef long off_t;

int ftruncate (FileRef fr, off_t length){
	Err err=VFSFileResize(fr,(UInt32) length);
	if(err=errNone) return 0;
	return -1;
}
/*
Char* toupper (Char *str)
{
	int i,j,k;
	Char c;
	j=StrLen(str);
	k='a'-'A';
	for(i=0;i<j;i++){
		c=*(str+i);
		if ((c>='a')&&(c<='z')) *(str+i)-=k;
	}
	return str;
}

Char* tolower (Char *str)
{
	int i,j,k;
	Char c;
	j=StrLen(str);
	k='a'-'A';
	for(i=0;i<j;i++){
		c=*(str+i);
		if ((c>='A')&&(c<='Z')) *(str+i)+=k;
	}
	return str;
}
*/

UInt32 iterator;
FileInfoType infoP;

FileRef opendir (const char *dirname){
	FileRef fr;
	iterator=vfsIteratorStart;
	Err err=VFSFileOpen(volumeRef,dirname,vfsModeReadWrite,&fr);
	if(err==errNone) return fr;
	return 0;
}

int closedir (FileRef fr){
	Err err=VFSFileClose(fr);
	return (int)err;
}

char name[128];
FileInfoType *readdir(FileRef fr){
	infoP.nameP=name;
	infoP.nameBufLen=128;
	Err err=VFSDirEntryEnumerate(fr,&iterator,&infoP);
	if (err!=errNone) return NULL;
	return &infoP;
}

int access(const char* path,int mode)
{
	FileRef fr;
	Err err=VFSFileOpen(volumeRef,path,vfsModeRead,&fr);
	if (err!=errNone) return -1;
	VFSFileClose(fr);
	return 0;
}

int rename (const char *old, const char *newname){
	return (int)VFSFileRename(volumeRef,old,newname);
}

FileRef fileno(FILE *fr) {return (FileRef)fr;}

int rmdir (const char *path){
	Err err=VFSFileDelete(volumeRef,path);
	if(err!=errNone) return -1;
	return 0;
}
int mkdir (const char *path, mode_t mode){
	Err err=VFSDirCreate(volumeRef,path);
	if(err!=errNone) return -1;
	return 0;
}

struct tm _time__;
struct tm* localtime(const time_t *clock){
	DateTimeType dtt;
	TimSecondsToDateTime(*clock,&dtt);
	_time__.tm_sec=dtt.second;
	_time__.tm_min=dtt.minute;
	_time__.tm_hour=dtt.hour;
	_time__.tm_mday=dtt.day;
	_time__.tm_mon=dtt.month;
	_time__.tm_year=dtt.year;
	_time__.tm_wday=dtt.weekDay;
	_time__.tm_yday=0;
	_time__.tm_isdst=0;
}
