/*
 *  Copyright (C) 2002-2006  The DOSBox Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* $Id: sdlmain.cpp,v 1.113 2006/03/29 12:26:07 qbix79 Exp $ */

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif

#include "dosbox.h"
#include "video.h"
#include "mouse.h"
#include "pic.h"
#include "timer.h"
#include "setup.h"
#include "support.h"
#include "debug.h"
#include "mapper.h"
#include "vga.h"
#include "keyboard.h"
#include "loader.h"

#define DISABLE_JOYSTICK

#if !(ENVIRON_INCLUDED)
extern char** environ;
#endif


#define STDOUT_FILE	TEXT("stdout.txt")
#define STDERR_FILE	TEXT("stderr.txt")
#define DEFAULT_CONFIG_FILE "dosbox.conf"

enum SCREEN_TYPES	{ 
	SCREEN_SURFACE,
	SCREEN_SURFACE_DDRAW,
	SCREEN_OVERLAY,
	SCREEN_OPENGL
};

enum PRIORITY_LEVELS {
	PRIORITY_LEVEL_LOWEST,
	PRIORITY_LEVEL_LOWER,
	PRIORITY_LEVEL_NORMAL,
	PRIORITY_LEVEL_HIGHER,
	PRIORITY_LEVEL_HIGHEST
};


struct SDL_Block {
	bool active;							//If this isn't set don't draw
	bool updating;
	struct {
		Bit32u width;
		Bit32u height;
		Bit32u bpp;
		Bitu flags;
		double scalex,scaley;
		GFX_CallBack_t callback;
	} draw;
	bool wait_on_error;
	struct {
		struct {
			Bit16u width, height;
			bool fixed;
		} full;
		struct {
			Bit16u width, height;
		} window;
		Bit8u bpp;
		bool fullscreen;
		bool doublebuf;
		SCREEN_TYPES type;
		SCREEN_TYPES want_type;
	} desktop;
	struct {
		PRIORITY_LEVELS focus;
		PRIORITY_LEVELS nofocus;
	} priority;
	struct {
		bool autolock;
		bool autoenable;
		bool requestlock;
		bool locked;
		Bitu sensitivity;
	} mouse;
};

extern char * RunningProgram;
extern bool CPU_CycleAuto;
//Globals for keyboard initialisation
bool startup_state_numlock=false;
bool startup_state_capslock=false;
void GFX_SetTitle(Bits cycles,Bits frameskip,bool paused){
	char title[200]={0};
	static Bits internal_cycles=0;
	static Bits internal_frameskip=0;
	if(cycles != -1) internal_cycles = cycles;
	if(frameskip != -1) internal_frameskip = frameskip;
	if(CPU_CycleAuto)
		sprintf(title,"DOSBox %s, Cpu Cycles:     auto, Frameskip %2d, Program: %8s",VERSION,internal_frameskip,RunningProgram);
	else
		sprintf(title,"DOSBox %s, Cpu Cycles: %8d, Frameskip %2d, Program: %8s",VERSION,internal_cycles,internal_frameskip,RunningProgram);

	if(paused) strcat(title," PAUSED");
}

static void PauseDOSBox(bool pressed) {
}
 

/* Reset the screen with current values in the sdl structure */
Bitu GFX_GetBestMode(Bitu flags) {
	Bitu testbpp,gotbpp;
	{
		/* Check if we can satisfy the depth it loves */
		if (flags & GFX_LOVE_8) testbpp=8;
		else if (flags & GFX_LOVE_15) testbpp=15;
		else if (flags & GFX_LOVE_16) testbpp=16;
		else if (flags & GFX_LOVE_32) testbpp=32;
		else gotbpp=16;
		/* If we can't get our favorite mode check for another working one */
		switch (gotbpp) {
		case 8:
			if (flags & GFX_CAN_8) flags&=~(GFX_CAN_15|GFX_CAN_16|GFX_CAN_32);
			break;
		case 15:
			if (flags & GFX_CAN_15) flags&=~(GFX_CAN_8|GFX_CAN_16|GFX_CAN_32);
			break;
		case 16:
			if (flags & GFX_CAN_16) flags&=~(GFX_CAN_8|GFX_CAN_15|GFX_CAN_32);
			break;
		case 24:
		case 32:
			if (flags & GFX_CAN_32) flags&=~(GFX_CAN_8|GFX_CAN_15|GFX_CAN_16);
			break;
		}
		flags |= GFX_CAN_RANDOM;
	}
	return flags;
}


void GFX_ResetScreen(void) {
	GFX_Stop();
	GFX_Start();
}


struct loader *loadf;

Bitu GFX_SetSize(Bitu width,Bitu height,Bitu flags,double scalex,double scaley,GFX_CallBack_t callback) {
	GFX_EndUpdate( 0 ,(BitmapType*)loadf->bmph);

	Bitu bpp;
	Bitu retFlags;
	
	{
		if (flags & GFX_CAN_8) bpp=8;
		if (flags & GFX_CAN_15) bpp=15;
		if (flags & GFX_CAN_16) bpp=16;
		if (flags & GFX_CAN_32) bpp=32;
		retFlags = GFX_CAN_16;
	}
	if (retFlags) 
		GFX_Start();
	return retFlags;
}

void GFX_CaptureMouse(void) {
}

bool mouselocked; //Global variable for mapper
static void CaptureMouse(bool pressed) {
	GFX_CaptureMouse();
}

void GFX_SwitchFullScreen(void) {
}

static void SwitchFullScreen(bool pressed) {
	GFX_SwitchFullScreen();
}


bool GFX_StartUpdate(Bit8u * & pixels,Bitu & pitch) {
	return true;
}

int x=0,y=0;
extern bool directscr;
void GFX_EndUpdate( const Bit16u *changedLines , BitmapType *bmph) {
  if(!directscr)WinDrawBitmap(bmph,x,y);
}

void GFX_SetPalette(Bitu start,Bitu count,GFX_PalEntry * entries) {
	WinPalette (winPaletteSet,start,count,(RGBColorType*)entries);
}

Bitu GFX_GetRGB(Bit8u red,Bit8u green,Bit8u blue) {
	Bit32u retval=((red<<8)&0xF800)|((green<<3)&0x07E0)|(blue>>3);
	if (!directscr)return ((retval>>8)|(retval<<8));
	return retval;
}

void GFX_Stop() {
}

void GFX_Start() {
}

static void GUI_ShutDown(Section * sec) {
	GFX_Stop();
}

static void KillSwitch(bool pressed) {
}

static void SetPriority(PRIORITY_LEVELS level) {
}

//static unsigned char logo[32*32*4]= {
//#include "dosbox_logo.h"
//};

static void GUI_StartUp(Section * sec) {
}

void Mouse_AutoLock(bool enable) {
}

KBD_KEYS realky (unsigned char *uch){
	if(!*uch)switch (*(uch+1)){
		case 'a': return KBD_a;
		case 'b': return KBD_b;
		case 'c': return KBD_c;
		case 'd': return KBD_d;
		case 'e': return KBD_e;
		case 'f': return KBD_f;
		case 'g': return KBD_g;
		case 'h': return KBD_h;
		case 'i': return KBD_i;
		case 'j': return KBD_j;
		case 'k': return KBD_k;
		case 'l': return KBD_l;
		case 'm': return KBD_m;
		case 'n': return KBD_n;
		case 'o': return KBD_o;
		case 'p': return KBD_p;
		case 'q': return KBD_q;
		case 'r': return KBD_r;
		case 's': return KBD_s;
		case 't': return KBD_t;
		case 'u': return KBD_u;
		case 'v': return KBD_v;
		case 'w': return KBD_w;
		case 'x': return KBD_x;
		case 'y': return KBD_y;
		case 'z': return KBD_z;
		case '-': return KBD_slash;
		case ',': return KBD_semicolon;
		case '.': return KBD_period;
		case 0x08: return KBD_backspace;
		case 0x09: return KBD_tab;
		case 0x20: return KBD_space;
		case 0x0D: return KBD_enter;
		case 0x1C: return KBD_left;
		case 0x1D: return KBD_right;
		case 0x1E: return KBD_up;
		case 0x1F: return KBD_right;
		default: return KBD_NONE;
	}
	else if(*uch==0x01) return KBD_capslock;
	else if(*uch==0xC0) switch (*(uch+1)){//red
		case 0x71: return KBD_f1;
		case 0x77: return KBD_f2;
		case 0x65: return KBD_f3;
		case 0x72: return KBD_f4;
		case 0x74: return KBD_f5;
		case 0x79: return KBD_f6;
		case 0x75: return KBD_f7;
		case 0x69: return KBD_f8;
		case 0x6F: return KBD_f9;
		case 0x70: return KBD_f10;
		default: return KBD_NONE;
	}
	else if(*uch==0x90) switch (*(uch+1)){//blue
		case 0x71: return KBD_1;
		case 0x77: return KBD_2;
		case 0x65: return KBD_3;
		case 0x72: return KBD_4;
		case 0x74: return KBD_5;
		case 0x79: return KBD_6;
		case 0x75: return KBD_7;
		case 0x69: return KBD_8;
		case 0x6F: return KBD_9;
		case 0x70: return KBD_0;
		default: return KBD_NONE;
	}
	else if(*uch==0x90) switch (*(uch+1)){
		case 0x71: return KBD_1;
		case 0x77: return KBD_2;
		case 0x65: return KBD_3;
		case 0x72: return KBD_4;
		case 0x74: return KBD_5;
		case 0x79: return KBD_6;
		case 0x75: return KBD_7;
		case 0x69: return KBD_8;
		case 0x6F: return KBD_9;
		case 0x70: return KBD_0;
		default: return KBD_NONE;
	}
	return KBD_NONE;
}
int oldx=0,oldy=0,newx,newy,mousemod=4;
bool pressed=false;
void MAPPER_CheckEvent(EventType * event);
void callm68k(void*,void*);
void GFX_Events() {
	EventType *e=loadf->e;
	unsigned char *ev;
	ev=(unsigned char*)e;
	callm68k(loadf->funcptr,NULL);
	unsigned int p=KeyCurrentState();
	if(p&0x00000002){y-=2;if(y<-40)y=-40;}
	else if(p&0x00000004){y+=2;if(y>0)y=0;}
	else if(p&0x00000010){x-=2;if(x<-160)x=-160;}
	else if(p&0x00000020){x+=2;if(x>0)x=0;}
	else if(p!=0){
		if(!pressed) Mouse_ButtonPressed(0);
		pressed=true;
	}
	else{
		if(pressed) Mouse_ButtonReleased(0);
		pressed=false;
	}
	switch(ev[1]){
		case 4: switch (ev[9]){
		   case 32: {// key down event
			KBD_KEYS kcode=(KBD_KEYS)realky(ev+10);
			if(!kcode) break;
			if(kcode==KBD_semicolon) KEYBOARD_AddKey(KBD_leftshift,true);
		   	KEYBOARD_AddKey(kcode,true);
		   	KEYBOARD_AddKey(kcode,false);
			if(kcode==KBD_semicolon) KEYBOARD_AddKey(KBD_leftshift,false);
		   }
		}
		break;
	 	case 3://penMoveEvent:
	 	if((ev[4])||(ev[6])) break;
		newx=ev[5]*mousemod;
		newy=ev[7]*mousemod;
		Mouse_CursorMoved((float)newx-oldx,(float)newy-oldy,(float)newx/159,(float)newy/159,true);
		oldx=newx;
		oldy=newy;
		break;
	}
}

/* static variable to show wether there is not a valid stdout.
 * Fixes some bugs when -noconsole is used in a read only directory */
static bool no_stdout = false;

void GFX_ShowMsg(char * format,...) {
};
#include "render_scalers.h"
void Delete_vectors();
void Init_vectors();
void Lang_Init();
void Lang_Destroy();
void Init_Auto();
void cleanup();
unsigned char *scalercache;
unsigned char *mem;
extern scalerSourceCache_t *scalerSourceCache;

CommandLine *com_line;
unsigned int memorysize;
int main(struct loader *loadf1) {
		loadf=loadf1;
		volumeRef=loadf->volume;
//		if(VFSFileOpen(volumeRef,"/PALM/1.log",vfsModeWrite|vfsModeCreate|vfsModeTruncate,&logfr)!=errNone) logfr=0;
		mem=loadf->mem;
		scalerSourceCache=(scalerSourceCache_t*)loadf->scalercache;
		memorysize=loadf->memsize;
		Init_Auto();
		char *linije[1]={"palmdosbox"};
		com_line=new CommandLine(1,linije);
		Config *myconf=new Config(com_line);
		control=myconf;
		DOSBOX_Init();
		Lang_Init();

		if (control->ParseConfigFile("/dosbox.conf") == false)  {
//					LOG_MSG("CONFIG: Using default settings. Create a configfile to change them");
			WinDrawChars("CONFIG: Using default settings.",31,0,0);
		}
		else WinDrawChars("CONFIG: Using non-default settings.",35,0,0);
		control->Init();
//		MAPPER_Init();
		control->StartUp();
		delete myconf;
		delete com_line;
		Lang_Destroy();
		cleanup();
		WinDrawChars("Exiting ...",11,0,0);
	return 0;
};
